from imutils.video import VideoStream
import cv2
import imutils
import time
import numpy as np
import pickle
import matplotlib.pyplot as plt

# untere und obere Grenzen der Farben im HSV Farbraum
green = (70, 50, 0 )
GREEN = (130, 255, 150)

red = (0, 150, 0)
RED = (100, 255, 255)

blue = (100, 100, 50)
BLUE = (150, 255, 255)

# Abmessungen des Kranarms
c = 700
b = 100

# frameSize ermitteln
vs = VideoStream(src=0).start()
time.sleep(2.0)
frame = vs.read()
vs.stop()
frame = imutils.resize(frame, width=600)
frameSize = (frame.shape[0], frame.shape[1])
del frame

##### Definitionen
def write (obj, name):             # Daten abspeichern
    fileObj = open(name, 'wb')
    pickle.dump(obj,fileObj)
    fileObj.close()

def read (name):                   # Daten auslesen
    fileObj = open(name, 'rb')
    obj = pickle.load(fileObj)
    fileObj.close()
    return obj

##### Felder in Koordinaten übersetzen
def mitte (eins, zwei, c):         # Mittelpunkt zwischen Punkten
    dx = zwei[0] - eins[0]
    dy = zwei[1] - eins[1]
    
    d = np.sqrt(dx**2 + dy**2)
    f = c * d/8
    e = (eins[0] + f * dx/d, eins[1] + f * dy/d)
    
    plt.scatter(e[0], e[1])         
    return e

def field2Coor (fields):           # Feld zu Koordinatenpunkte 
    for a in range(0,4):
        plt.scatter(src[a][0],src[a][1])
    
    x = fields[0]
    y = fields[1]
    
    def feld ():
        p1 = mitte(src[1], src[0], y)
        p3 = mitte(src[3], src[2], y)
        linkeEcke = mitte(p1, p3, x)
        return linkeEcke
        
    a = feld()
    x -= 1
    y -= 1
    b = feld()
    
    k = mitte(a, b, 4)
    plt.show()
    return k

##### Aktuelle Position der Greifklaue bestimmen
def getDot (lower, upper):         # Mittelpunkt und Radius von farbigem Punkt
    
    vs = VideoStream(src=0).start()
    time.sleep(2.0)
    
    frame = vs.read()
    vs.stop()
    frame = imutils.resize(frame, width=600)
    blurred = cv2.GaussianBlur(frame, (11, 11), 0)
    hsv = cv2.cvtColor(blurred, cv2.COLOR_BGR2HSV)

    mask = cv2.inRange(hsv, lower, upper)
    mask = cv2.erode(mask, None, iterations=2)
    mask = cv2.dilate(mask, None, iterations=2)
    
    cnts = cv2.findContours(mask.copy(), cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    cnts = imutils.grab_contours(cnts)
    center = None
    
    if len(cnts) > 0:
        c = max(cnts, key=cv2.contourArea)
        ((x, y), radius) = cv2.minEnclosingCircle(c)
        M = cv2.moments(c)
        center = (int(M["m10"] / M["m00"]), int(M["m01"] / M["m00"]))
        
        if radius > 10 and radius < 30: return center[0], center[1], radius
        else: 
            time.sleep(1)
            return getDot(lower, upper)

    else: 
        print("Keinen Punkt gefunden")
        return

def projection (point):            # Projektion auf Schachbrett, um Perspektive zu korrigieren
    middle = (int(frameSize[1]/2), int(frameSize[0]/2))
    dx = point[0] - middle[0]
    dy = point[1] - middle[1]
    d = np.sqrt(dx**2 + dy**2)
    
    dn = d * (1-b/c)
    
    if d != 0:  dx2 = (dn/d) * dx
    if dx != 0: dy2 = (dx2/dx) * dy
    else: dy2 = dn
    
    return (middle[0] + int(dx2), middle[1] + int(dy2))

def getPos ():                     # Position der Greifklaue ermitteln    
    posRed = getDot(red, RED)
    posBlue = getDot(blue, BLUE)
    
    if posRed != None and posBlue != None: # offset berechenen mit Rot blau
        px = 20
        n=[posRed[1]-posBlue[1],posBlue[0]-posRed[0]]
        l=np.sqrt(n[0]**2+n[1]**2)
        n[0]/=l
        n[1]/=l
        mid = ((posRed[0]+posBlue[0])/2, (posRed[1]+posBlue[1])/2)
        pnt = (mid[0]-px*n[0],mid[1]-px*n[1])
        return pnt
        
    elif posRed == None and posBlue != None: # offset berechnen mit Grün Blau        
        posGreen = getDot(green, GREEN)
        px = 95
        x = 6
        n=[posGreen[1]-posBlue[1],posBlue[0]-posGreen[0]]
        l=np.sqrt(n[0]**2+n[1]**2)
        n[0]/=l
        n[1]/=l
        mid = mitte(posGreen, posBlue, x)
        pnt = (mid[0]+px*n[0],mid[1]+px*n[1])
        return pnt
        
    elif posRed != None and posBlue == None: # offset berechnen mit Rot Grün
        posGreen = getDot(green, GREEN)
        px = 100
        x = 6.5
        n=[posGreen[1]-posRed[1],posRed[0]-posGreen[0]]
        l=np.sqrt(n[0]**2+n[1]**2)
        n[0]/=l
        n[1]/=l
        mid = mitte(posGreen, posRed, x)
        pnt = (mid[0]-px*n[0],mid[1]-px*n[1])
        return pnt
    
    raise Exception("Punkte nicht erkannt")

##### Arduino Schnittstelle initialisieren
import pyfirmata
board = pyfirmata.Arduino("COM8")

def stepper (pin, delay, steps, direc): # Stepper bewegen, mit direc: 0 = CCW, 1 = CW
    delay /= 2
    try:
        board.digital[pin].mode = pyfirmata.OUTPUT
        
    finally:
        if steps < 0:
            steps = -steps
        board.digital[pin-1].write(direc)

        for i in range(0,int(steps)):
            board.digital[pin].write(1)
            time.sleep(delay/1000)
            board.digital[pin].write(0)
            time.sleep(delay/1000)

def wagen (delay, steps, direct):       # Seilwinden gleichzeitig bedienen, um Wagen zu ziehen
     for i in range(steps):
         stepper(5, delay, 1, direct)
         stepper(7, delay, 3, direct)
    
servoPin = 8
serv = board.get_pin("d:"+str(servoPin)+":s")

def servo (deg):  serv.write(deg)
def servoOpen():  serv.write(25)
def servoClose(): serv.write(180)


##### Roboterarm lenken

# src laden
try:    src = read(r"C:\Users\Yannik Mika\OneDrive - Technische Universität Berlin\Robotik\Webcam Capture\src.obj")
except: src = np.float32([(50,   50), # links oben
                          (450,  50), # rechts oben
                          (50,  450), # links unten
                          (450, 450)])# rechts unten

def goto (coor):   
    while True:
        pos = projection(getPos())
        
        # Ausrichtung in y-Richtung        
        if   pos[1] > coor[1]: stepper(3, 1000, 50, 0)
        elif pos[1] < coor[1]: stepper(3, 1000, 50, 1)
        
        # Ausrichtung in x-Richtung
        if   pos[0] > coor[0]: wagen(1000, 100, 0)
        elif pos[0] < coor[0]: wagen(1000, 100, 1)
            
        # Abbruchbedingung
        if np.sqrt((pos[0]-coor[0])**2 + (pos[1]-coor[1])**2) < 10: return

fields = (6, 4)
coor = field2Coor(fields)
goto(coor)